#include "Sound.h"
#include <QString>
#include <QSound>
#include <QtDebug>

#if QT_VERSION > 0x050000
Sound::Sound(const QString& sound_file)
{
    mSound = new QSound(sound_file);
    mSound->setLoops(-1);
}

Sound::Sound(QObject *parent, const QString &sound_file) : QObject(parent)
{
    mSound = new QSound(sound_file, this);
    mSound->setLoops(-1);
}

Sound::~Sound()
{
    delete mSound;
}

void Sound::play()
{
    if(mSound->isFinished())
        mSound->play();
}

void Sound::stop()
{
    mSound->stop();
}

#else

Sound::Sound(QString sound_file)
{

    //create phonon objects
    audioOutput = new Phonon::AudioOutput(Phonon::MusicCategory, this);
    mediaObject = new Phonon::MediaObject(this);
    source =  Phonon::MediaSource(sound_file);
    connect(mediaObject, SIGNAL(finished()), this, SLOT(finished()));
    connect(&loop_timer, SIGNAL(timeout()), this, SLOT(timer_timeout()));
    audiograph = Phonon::createPath(mediaObject, audioOutput);
    mediaObject->setCurrentSource(source);
    set_loop_delay(500);
    loop_timer.setSingleShot(true);
    _loop = false;
}

void Sound::play()
{
    // release_19: play calls loop
    // in order to offer the same interface
    // as QT 5 version above

    // old
    /*
    if (_loop) return;
    mediaObject->play();
    */
    // new
    loop();
}

void Sound::loop()
{
    if(!_loop){
        _loop = true;
        mediaObject->play();
    }
}

void Sound::stop()
{
    _loop = false;
    loop_timer.stop();
    mediaObject->stop();

}

void Sound::finished()
{
    mediaObject->stop();
    loop_timer.setSingleShot(true);
    loop_timer.start();
}

void Sound::set_loop_delay(int delay)
{
    loop_timer.setInterval(delay);
    _delay=delay;
}

int  Sound::get_loop_delay()
{
    return loop_timer.interval();
}

void Sound::timer_timeout(){
    if (_loop){
         mediaObject->play();
    }
}


#endif
