/***************************************************************************
 *   Copyright (C) 2007 by Claudio Scafuri, Giacomo Strangolino   *
 *   claudio@hyo   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "alarm.h"
#include "filter.h"
#include "acmd.h"
#include "AlarmTreeWidget.h"
#include "AlarmTreeWidgetItem.h"
#include <QMessageBox>
#include <QCloseEvent>
#include "AlarmHistoryTree.h"
#include <QDebug>
#include <QAbstractItemModel>
#include <QDesktopServices> // open url

// cumbia
#include <qulogimpl.h>
#include <cucontrolsfactorypool.h>
#include <cumbiapool.h>
#include <quapps.h>
#include <quwriter.h>
#include <quwatcher.h>
#include <qustring.h>
#include <qustringlist.h>
#include <QTimer>
// cumbia

Alarm::Alarm(CumbiaPool *cu_p, QWidget *parent) : QWidget(parent)
{
    // cumbia
    CuModuleLoader mloader(cu_p, &m_ctrl_factory_pool, &m_log_impl);
    cu_pool = cu_p;
    ui = new Ui::Alarm;
    ui->setupUi(this, cu_p, m_ctrl_factory_pool);

    // mloader.modules() to get the list of loaded modules
    // cumbia

    QString soundFileName, alarmSysTestServerName;
    QStringList args = qApp->arguments();

    connect(ui->pBAckAll, SIGNAL(clicked()), this, SLOT(ackAll()));
    connect(ui->pBStopSound, SIGNAL(clicked()), this, SLOT(stopSound()));

    // view types
    ui->treeWidgetAlarms->setType(AlarmTreeWidget::AlarmType);
    ui->twDisabledAlarms->setType(AlarmTreeWidget::DisabledAlarmType);

    /* silence alarm */
    connect(ui->treeWidgetAlarms, SIGNAL(silence()), this, SLOT(silence()));
    connect(ui->treeWidgetAlarms, SIGNAL(silence()), this, SLOT(stopSound()));
    /* ack. Also uses QTWriter to impart StopNew command */
    connect(ui->treeWidgetAlarms, SIGNAL(ack()), this, SLOT(ack()));
    /* enable / disable */
    connect(ui->treeWidgetAlarms, SIGNAL(enableAlarm()), this, SLOT(enableAlarm()));
    connect(ui->twDisabledAlarms, SIGNAL(enableAlarm()), this, SLOT(enableAlarm()));
    connect(ui->treeWidgetAlarms, SIGNAL(disableAlarm()), this, SLOT(disableAlarm()));

    connect(ui->pBStopSound, SIGNAL(clicked()), this, SLOT(stopSound()));
    connect(ui->tabWidget, SIGNAL(currentChanged(int)), this, SLOT(tabChanged(int)));


    /* message label */
    ui->msgLabel->setVisible(false);
    ui->pBConfHistory->setHidden(true);

    // device list from Tango DB free prop
    QString devlist_prop;
    args.contains("--device-list-prop") && args.size() > args.indexOf("--device-list-prop") + 1
            ? devlist_prop = args.at(args.indexOf("--device-list-prop") + 1) : devlist_prop = "DeviceList";
    QuWatcher *db_dlist_w = new QuWatcher(this, cu_pool, m_ctrl_factory_pool); // device prop watcher
    db_dlist_w->setObjectName("free_prop_w");
    db_dlist_w->setSingleShot(true);
    db_dlist_w->setAutoDestroy(true);
    connect(db_dlist_w, SIGNAL(newData(CuData)), this, SLOT(onDevList(CuData)));
    db_dlist_w->setSource(QString("tango://#AlarmHandler#%1").arg(devlist_prop));

    // db params
    QuWatcher *db_conn_p = new QuWatcher(this, cu_pool, m_ctrl_factory_pool); // device prop watcher
    db_conn_p->setObjectName("db_conn_p");
    db_conn_p->setSingleShot(true);
    db_conn_p->setAutoDestroy(true);
    connect(db_conn_p, SIGNAL(newData(CuData)), ui->twHistory, SLOT(onDbParamsReady(CuData)));
    db_conn_p->setSource(QString("tango://#AlarmHandler#HistoryDb"));

    // groups
    QuWatcher *grpW = new QuWatcher(this, cu_pool, m_ctrl_factory_pool);
    grpW->setObjectName("group_w");
    grpW->setSingleShot(true);
    grpW->setAutoDestroy(true);
    connect(grpW, SIGNAL(newData(CuData)), this, SLOT(onGroupList(CuData)));
    grpW->setSource(QString("tango://AlarmHandler(GroupNames)").arg(devlist_prop));

    //---------------------------------------------------------------------------------------
    if(args.contains("--alarm-test-devname") && args.size() > args.indexOf("--alarm-test-devname") + 1) {
        alarmSysTestServerName = args.at(args.indexOf("--alarm-test-devname") + 1);
        // printf("\e[1;32m --alarm-test-devname specified: \"%s\"\e[0m\n", qstoc(alarmSysTestServerName));
        ui->tPBTestAlarm->setTarget(alarmSysTestServerName + "/alarm_test(true)");
        ui->tPBTestAlarmStop->setTarget(alarmSysTestServerName + "/alarm_test(false)");
    }
    else if(args.contains("--alarm-test-default-dev")) {
        printf("no --alarm-test-devname specified: using ecrfpsa/alarmtest/0/alarm_test\n");
        ui->tPBTestAlarm->setTarget("ecrfpsa/alarmtest/0/alarm_test(true)");
        ui->tPBTestAlarmStop->setTarget("ecrfpsa/alarmtest/0/alarm_test(false)");
    }
    //---------------------------------------------------------------------------------------

    if(args.contains("--sound-file-name") && args.size() > args.indexOf("--sound-file-name")) {
        soundFileName = args.at(args.indexOf("--sound-file-name") + 1);
        qDebug() <<"Sound file name: "<< soundFileName;
    }
    else
        printf("--> Alarm::Alarm: use \e[1;32m--sound-file-name path/to/sound_file_name.wav\e[0m to enable sounds\n");
    sound = new Sound(soundFileName);
    connect( ui->pBTestSoundOn,SIGNAL(clicked()), sound, SLOT(play()));
    connect( ui->pBTestSoundOff,SIGNAL(clicked()), sound, SLOT(stop()));

    connect(ui->cBStatus, SIGNAL(toggled(bool)), ui->comboStatus, SLOT(setEnabled(bool)));
    ui->pbInfo->setEnabled(false);
    ui->liViewInfo->setVisible(false);
    connect(ui->liViewInfo, SIGNAL(urlClicked(QString)), this, SLOT(openUrl(QString)));
    connect(ui->pbInfo, SIGNAL(toggled(bool)), this, SLOT(onInfoListViewVisibilityChange(bool)));

    foreach(QTreeWidget *tw, QList<QTreeWidget *>() << ui->treeWidgetAlarms << ui->twDisabledAlarms) {
        connect(tw, SIGNAL(message(QString)), ui->msgLabel, SLOT(setText(QString)));
        connect(tw, SIGNAL(mouseOverItem(bool)), ui->msgLabel, SLOT(setVisible(bool)));
        connect(tw, SIGNAL(itemSelectionChanged()), this, SLOT(onItemSelectionChange()));
        connect(tw, SIGNAL(doubleClicked(QModelIndex)), this, SLOT(onItemDoubleClicked(QModelIndex)));
    }
    /* connections for test section */
    //connect(ui->pBTestSoundOn, SIGNAL(clicked()), this, SLOT(testSound()));
    //connect(ui->pBTestSoundOff, SIGNAL(clicked()), this, SLOT(stopSound()));
    /* sound: beep on new alarm */
    connect(ui->treeWidgetAlarms, SIGNAL(newAlarmDetected()), this, SLOT(beep()));

    ui->tabWidget->setCurrentIndex(0);

    // database
    ui->dateStart->setDate(QDate::currentDate());
    ui->dateStop->setDate(QDate::currentDate());
    ui->timeStart->setTime(QTime::currentTime());
    ui->timeStop->setTime(QTime::currentTime());
    connect(ui->pBApplyHistory, SIGNAL(clicked()), this, SLOT(queryDB()));
    connect(ui->comboPeriod, SIGNAL(currentIndexChanged(int)), this, SLOT(enableFrameHistoryInterval(int)));
    connect(ui->leFilterName, SIGNAL(textChanged(QString)), this, SLOT(onFilterNameChanged(QString)));
    connect(ui->comboStatus, SIGNAL(currentTextChanged(QString)), this, SLOT(onFilterStatusChanged(QString)));
    connect(ui->cBStatus, SIGNAL(toggled(bool)), this, SLOT(onFilterStatusEnabled(bool)));
    connect(ui->pbClearFilter, SIGNAL(clicked()), ui->leFilterName, SLOT(clear()));

    m_filter = new AlarmFilter(ui->twHistory);
}

Alarm::~Alarm() {
    if(sound) {
        sound->stop();
        delete sound;
    }
    delete m_filter;
}

void Alarm::refresh(const CuData &da) {
    const QString& att = QuString(da, "src").section('/', -1, -1);
    AlarmTreeWidget *tw = att == "alarm" ? ui->treeWidgetAlarms : ui->twDisabledAlarms;
    tw->refresh(da);
    QString tn = "%1 [%2]";
    if(att == "alarm") ui->tabWidget->setTabText(0, tn.arg("Alarm").arg(tw->countAlarmed()));
    else ui->tabWidget->setTabText(1, tn.arg("Disabled Alarms").arg(tw->countAll()));
    // filter
    m_filter->filter();
}

void Alarm::onDevList(const CuData &da) {
    if(da["err"].toBool()) {
        perr("Alarm.onDevList: error fetching device list from database: src: \"%s\" msg: \"%s\"", vtoc2(da, "src"), vtoc2(da, "msg"));
    }
    else {
        m_handlers = QuStringList(da, "value");
        m_connect_devs(m_handlers);
    }
}

void Alarm::onGroupList(const CuData &da) {
    if(!da["err"].toBool() && da.containsKey("value")) {
//        foreach(QString g, QuStringList(da, "value"))
//            new QTreeWidgetItem(ui->tWGroup, QStringList() << g.remove("gr_"));
    } else
        perr("Alarm.onGroupList: error fetching groups from %s: %s", vtoc2(da, "src"), vtoc2(da, "msg"));
}

void Alarm::m_connect_devs(const QStringList &devs) {
    foreach(const QString& att, QStringList() << "alarm" << "alarmDisabled") {
        foreach(const QString dev, devs) {
            QuWatcher *alrw = new QuWatcher(this, cu_pool, m_ctrl_factory_pool);
            alrw->setObjectName(dev + "/alarm" + "+watcher");
            connect(alrw, SIGNAL(newData(CuData)), this, SLOT(refresh(CuData)));
            alrw->setSource(dev + "/" + att);
        }
    }
}

QList<QTreeWidgetItem *> Alarm::mGetSelectedItems()
{
    QList<QTreeWidgetItem *> selItems;
    /* get the tree widget that is currently visible */
    if(ui->treeWidgetAlarms->isVisible())
        selItems = ui->treeWidgetAlarms->selectedItems();
    return selItems;
}

void Alarm::silence()
{
    Acmd acmd;
    QStringList errdevs; // list of "server error" devs
    acmd.silence(ui->treeWidgetAlarms, false, cu_pool, m_ctrl_factory_pool, errdevs);
    foreach(const QString& ed, errdevs)
        ui->treeWidgetAlarms->setServerErrNotNew(ed);
    sound->stop();
}

void Alarm::enableAlarm() {
    // when tab changes, selection is cleared
    // we'll either have selected alarms from disabled view (which are all
    // eligible for enabling) or *silenced* alarms from the active alarms
    // view (which are the only category eligible to be enabled back again)
    Acmd().enableAlarm(ui->treeWidgetAlarms, cu_pool, m_ctrl_factory_pool);
    Acmd().enableAlarm(ui->twDisabledAlarms, cu_pool, m_ctrl_factory_pool);
}

void Alarm::disableAlarm() {
    Acmd().disableAlarm(ui->treeWidgetAlarms, cu_pool, m_ctrl_factory_pool);
}

void Alarm::testSound() {
    sound->play();
}

void Alarm::beep() {
    showNormal();
    sound->play();
}

void Alarm::openUrl(const QString &u) const {
    QDesktopServices::openUrl(QUrl(u));
}

void Alarm::stopSound() {
    Acmd acmd;
    QStringList errdevs; // list of "server error" devs
    acmd.stop_new(ui->treeWidgetAlarms, true, cu_pool, m_ctrl_factory_pool, errdevs);
    foreach(const QString& ed, errdevs)
        ui->treeWidgetAlarms->setServerErrNotNew(ed);
    sound->stop();
}

void Alarm::ack() {
    Acmd acmd;
    QStringList errdevs; // list of "server error" devs
    acmd.ack(ui->treeWidgetAlarms, false, cu_pool, m_ctrl_factory_pool, errdevs);
    foreach(const QString& ed, errdevs) {
        ui->treeWidgetAlarms->setServerErrNotNew(ed);
        ui->treeWidgetAlarms->setServerErrAck(ed);
    }
}

void Alarm::ackAll() {
    Acmd acmd;
    QStringList errdevs; // list of "server error" devs
    acmd.ack(ui->treeWidgetAlarms, true, cu_pool, m_ctrl_factory_pool, errdevs);
    foreach(const QString& ed, errdevs) {
        ui->treeWidgetAlarms->setServerErrNotNew(ed);
        ui->treeWidgetAlarms->setServerErrAck(ed);
    }
}

void Alarm::onInfoListViewVisibilityChange(bool vis) {
    if(!vis)
        ui->liViewInfo->activate("", "");
    else {
        AlarmTreeWidget *atw = ui->tabWidget->currentIndex() == 0 ? ui->treeWidgetAlarms : ui->twDisabledAlarms;
        atw->selectedItems().size() == 1 ?
                    ui->liViewInfo->activate(atw->selectedItems().first()->data(AlarmTreeWidgetItem::Device, Qt::UserRole).toString(),
                                             atw->selectedItems().first()->text(AlarmTreeWidgetItem::Alarm))
                  : ui->liViewInfo->activate("", "");
    }
}

void Alarm::onItemSelectionChange() {
    QList<QTreeWidgetItem *> sits = qobject_cast<AlarmTreeWidget *>(sender())->selectedItems();
    QTreeWidgetItem *it = nullptr;
    if(sits.size() == 1)
        it = sits.first();
    ui->pbInfo->setEnabled(it != nullptr);
    if(ui->liViewInfo->isVisible() && it)
        ui->liViewInfo->activate(it->data(AlarmTreeWidgetItem::Device, Qt::UserRole).toString(), it->text(AlarmTreeWidgetItem::Alarm));
}

void Alarm::onItemDoubleClicked(const QModelIndex &mi) {
    ui->pbInfo->setChecked(true);
}

void Alarm::onFilterNameChanged(const QString &n) {
    n.isEmpty() ? m_filter->rem("Alarm") : m_filter->add("Alarm", n);
    m_filter->filter();
}

void Alarm::onFilterStatusChanged(const QString &s) {
    s.isEmpty() ? m_filter->rem("Status") : m_filter->add("Status", s);
    m_filter->filter();
}

void Alarm::onFilterStatusEnabled(bool en) {
    onFilterStatusChanged(en ? ui->comboStatus->currentText() : "");
}

void Alarm::tabChanged(int t) {
    // prevent issues with operations on selected alarms on the previous tab
    foreach(AlarmTreeWidget *a, findChildren<AlarmTreeWidget *>())
        a->deselectAll();
    if(t == 2) {
        QApplication::setOverrideCursor(Qt::WaitCursor);
        QTimer::singleShot(500, this, SLOT(queryDB()));
    }
    if(t > 1) {
        ui->pBAckAll->setHidden(true);
        ui->pBConfHistory->setHidden(false);
        ui->pBConfHistory->setChecked(true);
        ui->pBConfHistory->setHidden(true);
        ui->pBStopSound->setHidden(true);
    }
    else if(t < 2) {
        ui->pBStopSound->setHidden(false);
        ui->pBAckAll->setHidden(false);
        ui->pBConfHistory->setHidden(true);
        onInfoListViewVisibilityChange(ui->liViewInfo->isVisible());
    }
}

void Alarm::closeEvent(QCloseEvent *e)
{
    int reply = QMessageBox::question(this, "Warning", "Are you really sure you want to close this application?\n"
                                                       "If you do, you will not receive alarm notifications",
                                      QMessageBox::Yes|QMessageBox::No);

    if(reply == QMessageBox::Yes)
        reply = QMessageBox::question(this, "Warning", "By closing this panel, I am aware that I will not receive"
                                                       "\nalarms from now on",
                                      QMessageBox::Yes|QMessageBox::Cancel);
    if(reply == QMessageBox::Yes)
        e->accept();

    else
        e->ignore();

}

void Alarm::queryDB()
{
    QDateTime dts = QDateTime::currentDateTime();
    QString query;
    QDateTime dte(dts);
    switch(ui->comboPeriod->currentIndex())  {
    case 0:
        dts = dts.addDays(-1);
        dte = dte.addDays(1); // somewhere in the future
        break;
    case 1:
        dts = dts.addDays(-7);
        dte = dte.addDays(1); // somewhere in the future
        break;
    case 2:
        dts = dts.addMonths(-1);
        dte = dte.addDays(1); // somewhere in the future
        break;
    case 3:
        dts.setDate(ui->dateStart->date());
        dts.setTime(ui->timeStart->time());
        dte.setDate(ui->dateStop->date());
        dte.setTime(ui->timeStop->time());
        break;
    }

    QString ored_handlers;
    int i = 0;
    if(m_handlers.size() > 0)
        ored_handlers = "att_name LIKE '%" + m_handlers[0] + "' ";
    for(i = 1; i < m_handlers.size(); i++)
        ored_handlers += " OR att_name LIKE '%" + m_handlers[i] + "%' ";
    if(dts.isValid()  && dte.isValid() && dts < dte && ored_handlers.length() > 0) {
        query = QString("SELECT asd.data_time AS 'Date/Time',ac.att_name AS 'Alarm', "
        "COALESCE(REPLACE(REPLACE(REPLACE(REPLACE(REPLACE(REPLACE(asd.value_r,'0','NORMAL'),'1','ALARM UNACK'),'2','ALARM ACKED'),'3','RETURNED NORMAL'),'4','SHELVED'),'6','DISABLED'),aed.error_desc)  AS 'Status' "
        "FROM att_scalar_devenum_ro asd "
        "JOIN att_conf ac ON asd.att_conf_id=ac.att_conf_id "
        "LEFT OUTER JOIN att_error_desc aed ON aed.att_error_desc_id=asd.att_error_desc_id "
        "WHERE ac.att_name in (SELECT att_name FROM att_conf WHERE %1) "
        "AND asd.data_time > '%2' AND asd.data_time < '%3' ORDER BY asd.data_time DESC;").arg(ored_handlers).arg(dts.toString("yyyy-MM-dd hh:mm:ss.zzz")).arg(dte.toString("yyyy-MM-dd hh:mm:ss.zzz"));
    }
    else if(dts.isValid()  && dte.isValid() && dts >= dte)
        QMessageBox::information(this, "Invalid date / time range",  "Stop date/time is ahead of start date/time");
    else if(ored_handlers.isEmpty())
        QMessageBox::information(this, "No handlers", "No alarm handlers");

    if(query.length() > 0)
        ui->twHistory->executeQuery(query);

    printf("executing query \"%s\"\n", qstoc(query));

    m_filter->filter(true);
    QApplication::restoreOverrideCursor();
}

void Alarm::enableFrameHistoryInterval(int i) {
    ui->widgetHistoryConf->setEnabled(i == 3);
}
